/// \file a429_cdev.c
/// \brief инфраструктура для работы с символьными устройствами каналов pci устройства
#include "a429_cdev.h"
#include "a429_log.h"
#include "a429_regs.h"
#include "a429_info.h"
#include "a429_ioctl.h"

#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/fs.h>

/// \brief открытие символьного устройтва (канала)
static int a429cdev_open(struct inode *inode, struct file *file);
/// \brief закрытие символьного устройства (канала)
static int a429cdev_release(struct inode *inode, struct file *file);
/// \brief обработка команд символьного устройтсва (канала)
static long a429cdev_ioctl(struct file *file, unsigned int cmd, unsigned long arg);
/// \brief чтение из символьного устройтсва (канала)
static ssize_t a429cdev_read(struct file *file, char __user *buf, size_t count, loff_t *offset);
/// \brief запись в символьное устройтсво (канал)
static ssize_t a429cdev_write(struct file *file, const char __user *buf, size_t count, loff_t *offset);
/// \brief создание класса символьного устройтсва (канала)
void create_cdev_class(struct A429_DEV* driver_priv);
/// \brief уничтожение класса символьного устройтсва (канала)
void destroy_cdev_class(struct A429_DEV *driver_priv);

/// \brief таблица операций над каналами (имвольное устройство)
static const struct file_operations a429cdev_fops = {
    .owner              = THIS_MODULE,
    .open               = a429cdev_open,
    .release            = a429cdev_release,
    .unlocked_ioctl     = a429cdev_ioctl,
    .read               = a429cdev_read,
    .write              = a429cdev_write
};

/// \brief обрабочик события символьного устройтсва (канала)
/// установление прав - разрешение для юзерспейс
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 0, 0)
static int a429cdev_uevent(struct device *dev, struct kobj_uevent_env *env);
#else
static int a429cdev_uevent(const struct device *dev, struct kobj_uevent_env *env);
#endif
/// \brief деинициализация данных о символьном устройстве (канале)
static void deinit_cdev_data(struct a429_cdev_data* data);
/// \brief получить индекс символьного устройства (канала) из глобального массива символьных устройств cdevs
static int getCdevIndex(struct inode* inode);

/// \brief cdevs глобальный массив символьных устройств (каналов)
static struct a429_cdev_data cdevs[MAX_COUNT_BOARDS * (MAX_RECEIVE_CHANNELS + MAX_TRANSMIT_CHANNELS)];

static int getCdevIndex(struct inode* inode) {
    int i;  // счётчик
    uint32_t major, minor;  // мажор и минор идентификатор
    major = imajor(inode);
    minor = iminor(inode);
    for ( i = 0; i < (MAX_COUNT_BOARDS * MAX_CHANNELS); i++ ) {
        if ((cdevs[i].dev_major == major) && (cdevs[i].numCh == minor)) { // по уникальному major и minor получаем индекс из глобального массива
            return i;
        }
    }
    return -1;
}

struct a429_cdev_data* getCdevPtr(struct a429_cdev_private* priv) {
    if (priv->index == -1) // по индексу получаем ссылку на информационную структуру символьного устройтсва (канала)
        return NULL;
    else
        return &cdevs[priv->index];
}

struct a429_cdev_data* getCdevPtrByIndex(int index) {
    if (index == -1) // по индексу получаем ссылку на информационную структуру символьного устройтсва (канала)
        return NULL;
    else
        return &cdevs[index];
}

static void deinit_cdev_data(struct a429_cdev_data *data) {
    int j;  // счётчик
    data->dev_major = 0;
    data->drv = NULL;
    for ( j = 0; j < COUNT_NAME_SYMBOLS; j++ )
        data->devname[j] = 0;
    data->index = 0;
    data->numCh = 0;
}

void create_cdev_class(struct A429_DEV *driver_priv) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0)
    driver_priv->a429cdev_class = class_create(THIS_MODULE, driver_priv->a429cdev_name);
#else
    driver_priv->a429cdev_class = class_create(driver_priv->a429cdev_name);
#endif
    driver_priv->a429cdev_class->dev_uevent = a429cdev_uevent;
}

void destroy_cdev_class(struct A429_DEV* driver_priv) {
    class_unregister(driver_priv->a429cdev_class);
    class_destroy(driver_priv->a429cdev_class);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 0, 0)
static int a429cdev_uevent(struct device *dev, struct kobj_uevent_env *env) {
#else
static int a429cdev_uevent(const struct device *dev, struct kobj_uevent_env *env) {
#endif
    add_uevent_var(env, "DEVMODE=%#o", 0666);   // даём доступ user'у на запись/чтение
    return 0;
}

int create_char_devs(struct A429_DEV* driver_priv) {
    int i;              // счётчик
    int err;            // код ошибки
    int curNumCdev;     // текущий локальный номер символьного устройства (канала)
    dev_t dev;          // символьное устройство
    uint32_t major;     // мажор идентификатор
    snprintf(driver_priv->a429cdev_name, sizeof(driver_priv->a429cdev_name), "%s-%d-ch" , A429_CDEV_CLASS, driver_priv->unit); // генерим шаблонное имя для группы устройств
    err = alloc_chrdev_region(&dev, 0, driver_priv->curCountReceiveChannels + driver_priv->curCountTransmitChannels, driver_priv->a429cdev_name);  // выделяем место под заданное кол-во символьных устройтсв (каналов)
    if (err) {
        LOG_ERROR("create_char_devs err: %d", err);
        return err;
    }   
    major = MAJOR(dev);
    create_cdev_class(driver_priv); // создаём класс устройств с именем из a429cdev_name
    for (i = 0; i < MAX_CHANNELS; i++) {
        driver_priv->index_cdev[i] = -1;
    }
    curNumCdev = 0;
    for ( i = 0; i < (MAX_COUNT_BOARDS * MAX_CHANNELS); i++) { // обходим глобальный массив символьных устройств (каналов), чтобы найти свободное место
        if (curNumCdev < driver_priv->curCountReceiveChannels + driver_priv->curCountTransmitChannels) { //
            if (cdevs[i].drv != NULL)   // место - занято
                continue;
            else {  // место - свободно
                cdevs[i].index = i;
                cdevs[i].dev_major = major;
                cdevs[i].numCh = curNumCdev;
                cdevs[i].drv = driver_priv;
                cdev_init(&cdevs[i].cdev, &a429cdev_fops);
                cdevs[i].cdev.owner = THIS_MODULE;
                cdev_add(&cdevs[i].cdev, MKDEV(cdevs[i].dev_major, cdevs[i].numCh), 1);
                snprintf(cdevs[i].devname, sizeof(cdevs[i].devname), "%s-%d" , driver_priv->a429cdev_name, cdevs[i].numCh);  // генерим имя канала
                cdevs[i].device = device_create(driver_priv->a429cdev_class, NULL, MKDEV(cdevs[i].dev_major, cdevs[i].numCh), NULL, cdevs[i].devname); // создаём символьное устройтсво (канал)
                spin_lock_init(&cdevs[i].spinlockCh);
                driver_priv->index_cdev[curNumCdev] = i;
                LOG_RELEASE_INFO("cdev create: %s - index: %d", cdevs[i].devname, driver_priv->index_cdev[curNumCdev]);
                curNumCdev++;
                cdevs[i].numRkIn = RKIN_NONE;
            }
        }
    }
    return 0;
}

int destroy_char_devs(struct A429_DEV* driver_priv)
{
    int i;              // счётчик
    int dev_major;      // мажор идентификатор
    dev_major = BAD_DEV_MAJOR;
    for ( i = 0; i < (MAX_COUNT_BOARDS * MAX_CHANNELS); i++) { // обходим глобальный массив символьных устройств (каналов)
        if (driver_priv == cdevs[i].drv) {  // заданный канал - уничтожаем
            LOG_RELEASE_INFO("cdev destroy: %s", cdevs[i].devname);
            dev_major = cdevs[i].dev_major;
            device_destroy(driver_priv->a429cdev_class, MKDEV(dev_major, cdevs[i].numCh));
            deinit_cdev_data(&cdevs[i]);
        }
    }
    if (BAD_DEV_MAJOR == dev_major)
        return -1;
    destroy_cdev_class(driver_priv);    // уничтожаем класс
    unregister_chrdev_region(MKDEV(dev_major, 0), MINORMASK);   // разрегистриуем группу устройств
    LOG_RELEASE_INFO("destroy_char_devs is good!");
    return 0;
}

static int a429cdev_open(struct inode *inode, struct file *file)
{
    struct a429_cdev_private* priv;  // приватая структура символьного устройства
    priv = kzalloc(sizeof(struct a429_cdev_private), GFP_KERNEL);
    priv->index = getCdevIndex(inode);
    file->private_data = priv;
    return 0;
}

static int a429cdev_release(struct inode *inode, struct file *file)
{
    kfree((struct a429_cdev_data*)file->private_data);
    return 0;
}

static long a429cdev_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    long retVal;
    struct a429_cdev_data* dev = getCdevPtr((struct a429_cdev_private*)file->private_data);
    if (dev == NULL) {
        LOG_ERROR("a429cdev_ioctl - NULL");
        return -1;
    }
    retVal = ioctl_handler(dev, cmd, arg); // обработка ioctl-команд
    return retVal;
}

static ssize_t a429cdev_read(struct file *file, char __user *buf, size_t count, loff_t *offset)
{
    // struct mil1553_cdev_data* dev = getCdevPtr((struct mil1553_cdev_private*)file->private_data);
    // todo
    return -EPERM; // Operation not permitted
}

static ssize_t a429cdev_write(struct file *file, const char __user *buf, size_t count, loff_t *offset)
{
    // struct mil1553_cdev_data* dev = getCdevPtr((struct mil1553_cdev_private*)file->private_data);
    // todo
    return -EPERM; // Operation not permitted
}

T_INFO_CHANNEL get_info_channel(struct a429_cdev_data *data) {
    T_INFO_CHANNEL info;
    info.numberChannel = T_INFO_CHANNEL_BAD_VALUE;
    info.typeChannel = T_INFO_CHANNEL_BAD_VALUE;
    // определение типа канала
    if ((data->numCh >= 0) && (data->numCh < data->drv->curCountReceiveChannels)) {
        info.typeChannel = T_INFO_CHANNEL_TYPE_RECIEVER;
    } else if ((data->drv->curCountReceiveChannels >= 0) && (data->numCh < (data->drv->curCountReceiveChannels + data->drv->curCountTransmitChannels) )) {
        info.typeChannel = T_INFO_CHANNEL_TYPE_TRANSMITTER;
    } else
        return info;
    // определение номера передатчика/приёмника
    if ((data->numCh >= 0) && (data->numCh < data->drv->curCountReceiveChannels)) {
        info.numberChannel = data->numCh;
    } else if ((data->drv->curCountReceiveChannels >= 0) && (data->numCh < (data->drv->curCountReceiveChannels + data->drv->curCountTransmitChannels) )) {
        info.numberChannel = data->numCh - data->drv->curCountReceiveChannels;
    }
    return info;
}


void clear_cdev_array()
{
    int i;
    for (i=0; i<MAX_COUNT_BOARDS * (MAX_RECEIVE_CHANNELS + MAX_TRANSMIT_CHANNELS); i++) {
        cdevs[i].index = -1;
    }
}
