/// \file a429_ioctl.h
/// \brief инфраструктура ioctl-команд
#ifndef A429_CMD_H
#define A429_CMD_H

/// /// /// /// /// /// /// /// /// ///
/// \remark ioctl-команды драйвера
/// /// /// /// /// /// /// /// /// ///

#define IOC_MAGIC 'd'

/// /// /// /// ///
/// \remark IOCTL - ЧТЕНИЕ/ЗАПИСЬ из/в регистры
/// /// /// /// ///

/// \brief команда "запись в регистр"
/// позволяет записть значение в заданный регистр управления
#define IOCTL_WRITE_REG                             _IOW(IOC_MAGIC, 0,  SADDR_DATA)
/// \brief команда "запись в регистр заданных бит"
/// позволяет изменить отдельные биты в заданном регистре управления
#define IOCTL_WRITE_REG_BIT_MASK                    _IOW(IOC_MAGIC, 1,  SADDR_DATA_BIT_MASK)
/// \brief команда "чтение из регистра"
/// позволяет прочитать значение из заданного регистра управления
#define IOCTL_READ_REG                              _IOWR(IOC_MAGIC,2,  SADDR_DATA)

/// /// /// /// ///
/// \remark IOCTL - ЧТЕНИЕ ДАННЫХ из DMA БУФЕРА
/// /// /// /// ///

/// \brief команда "чтение из дма"
/// позволяет прочитать данные из дма
#define IOCTL_READ_DMA                              _IOWR(IOC_MAGIC,3,  T_CONTAINER_DMA)

/// \brief команда "сброс указателя dma канала"
/// позволяет обнулить DMA_INDEX и программные указатели чтения/записи
#define IOCTL_CLEAR_DMA                             _IO(IOC_MAGIC,  13)

/// /// /// /// ///
/// \remark IOCTL - Получение сервисной информации о драйвере и плате
/// /// /// /// ///

/// \brief команда "чтение подробной информации о плате и драйвере"
/// позволяет считать подробную информацию о pci-плате
#define IOCTL_VERSION                               _IOR(IOC_MAGIC, 4, VERSION)
/// \brief команда "чтение версии драйвера"
/// позволяет считать версию и дату драйвера
#define IOCTL_VERSION_DRIVER                        _IOR(IOC_MAGIC, 5, unsigned int)
/// \brief команда "чтение информации о pci-локации платы"
/// позволяет считать информацию pci
#define IOCTL_PCI_LOCATION                          _IOR(IOC_MAGIC, 6, PCI_LOCATION)

/// /// /// /// ///
/// \remark IOCTL - Взаимодействие с RAM передатчиков
/// /// /// /// ///

/// \brief команда "запись блока в RAM"
#define IOCTL_WRITE_TO_RAM                          _IOW(IOC_MAGIC, 7, RAM_BLOCK)
/// \brief команда "чтение блока из RAM"
#define IOCTL_READ_FROM_RAM                         _IOWR(IOC_MAGIC,8, RAM_BLOCK)

/// /// /// /// ///
/// \remark IOCTL - Инфоормация о канале
/// /// /// /// ///

/// \brief команда "информация о канале"
#define IOCTL_INFO_CHANNEL                          _IOWR(IOC_MAGIC, 9, T_INFO_CHANNEL)
/// \brief команда "информация о плате"
#define IOCTL_INFO_DEVICE                           _IOWR(IOC_MAGIC, 10, T_INFO_DEVICE)

/// /// /// /// ///
/// \remark IOCTL - Работа с прерываниями
/// /// /// /// ///

/// \brief команда "слежение за прерываниями"
#define IOCTL_TRACKING_INT                          _IOWR(IOC_MAGIC, 11, T_TRACKING_INT)
/// \brief команда "проверка информации слежения за прерываниями"
#define IOCTL_CHECK_TRACKING_INT                    _IOWR(IOC_MAGIC, 12, T_CHECK_TRACKING_INT)
/// \brief команда "запус канала в режиме передача по запросу"
#define IOCTL_START_AS_RK_IN                        _IOWR(IOC_MAGIC, 14, unsigned int)

/****************************************************************************************************/

/// /// /// /// /// /// /// /// /// ///
/// \remark типы данных для ioctl-команд драйвера
/// /// /// /// /// /// /// /// /// ///

/// /// /// /// ///
/// \remark DATA - ЧТЕНИЕ/ЗАПИСЬ из/в регистры
/// /// /// /// ///

/// \brief запись/чтение регистров
typedef struct {
    unsigned long daddr;        ///< адрес регистра (смещение в соотв. со спецификацией)
    unsigned int data;          ///< значение регистра
} SADDR_DATA;

/// \brief запись в регистр заданных бит
typedef struct {
    unsigned long daddr;        ///< адрес регистра (смещение в соотв. со спецификацией)
    unsigned int data;          ///< значение регистр
    unsigned int mask;          ///< битовая маска (1 - бит записывается из data, 0 - бит остаётся неизменным)
} SADDR_DATA_BIT_MASK;

/// /// /// /// ///
/// \remark DATA - ЧТЕНИЕ ДАННЫХ из DMA БУФЕРА
/// /// /// /// ///

/// \brief максимальное кол-во блоков дма в контейнере
#define DMA_COUNT_BLOCKS                    256
/// \brief размер одного блока дма в байтах
#define DMA_RAW_BLOCK_SIZE                  16
/// \brief размер дма буфера
#define DMA_SIZE_BYTES                      1048576

#pragma pack(push, 1)
/// \brief блок данных DMA
typedef struct {
    unsigned int word1;         ///< слово 1
    unsigned int word2;         ///< слово 2
    unsigned int word3;         ///< слово 3
    unsigned int word4;         ///< слово 4
} T_BLOCK_DMA;

/// \brief получить тип канала из слова дма 1
#define T_BLOCK_DMA_TYPE_CH(word1)  ((word1 >> 31) & 0x1)
/// \brief получить номер канала из слова дма 1
#define T_BLOCK_DMA_NUM_CH(word1)   ((word1 >> 24) & 0xF)

/// \brief контейнер DMA
typedef struct {
    unsigned int count;                         ///< кол-во блоков
    T_BLOCK_DMA blocks[DMA_COUNT_BLOCKS];       ///< блоки дма
} T_CONTAINER_DMA;
#pragma pack(pop)

/// /// /// /// ///
/// \remark DATA - Получение сервисной информации о драйвере и плате
/// /// /// /// ///

/// \brief размер строки pci-локации
#define SIZE_OF_PCI_LOC         13

/// \brief информация о локации на шине pci
typedef struct {
    ///
    /// \brief name инфо о локации
    /// например, 0000:00:02.0 -
    /// выделены домен (16 бит), шина (8 бит), устройство (5 бит) и функция (3 бита)
    ///
    char name[SIZE_OF_PCI_LOC];
} PCI_LOCATION;

/// \brief информация о драйвере и устройстве
typedef struct {
    unsigned int device_id;     ///< идентификатор устройства
    unsigned int vendor_id;     ///< вендор устройстваf
    unsigned int type;          ///< тип устройства (кол-во каналов)
    char revision;              ///< ревизия устройства
    char dev_name[30];          ///< имя символьного устройства
    int minor;                  ///< значение минора
    int irq;                    ///< номер прерывания
    long size_dma;              ///< размер ДМА буфера
    void* addr_dma_virt;        ///< виртуальный адрес ДМА буфера
    unsigned int pci_bars;      ///< адрес bar-пространства
    void* addr_bar_virt;        ///< виртуальный адрес bar-пространства
} VERSION;

/// /// /// /// ///
/// \remark DATA - Взаимодействие с RAM передатчиков
/// /// /// /// ///

/// \brief RAM блок - данные
#define RAM_BLOCK_TYPE_DATA             1
/// \brief RAM блок - дескрипторы
#define RAM_BLOCK_TYPE_DESCRIPTOR       2

/// \brief блок данных для записи в RAM передатчика
typedef struct {
    unsigned int type;      ///< тип инструкций (DATA - 1, DESCRIPTOR - 2)
    unsigned int shift;     ///< смещение от начала буфера в 32-х разрядных словах
    unsigned int length;    ///< размер поля dwords в 32-х разрядных словах
    unsigned int* dwords;   ///< данные (массив 32-х разрядных слов)
} RAM_BLOCK;

/// /// /// /// ///
/// \remark DATA - Информация о канале
/// /// /// /// ///

/// \brief информация об плате
typedef struct {
    int countRecieverChannels;          ///< кол-во приёмников
    int countTransmitterChannels;       ///< кол-во передатчиков
} T_INFO_DEVICE;

/// \brief тип канала - приёмник
#define T_INFO_CHANNEL_TYPE_RECIEVER        0
/// \brief тип канала - передатчик
#define T_INFO_CHANNEL_TYPE_TRANSMITTER     1
/// \brief тип канала - невалидное значение
#define T_INFO_CHANNEL_BAD_VALUE            -1

/// \brief The T_INFO_CHANNEL struct информация о канале
typedef struct {
    int typeChannel;        /// \brief тип канала (0 - приёмник, 1 - передатчик)
    int numberChannel;      /// \brief номер приёмника/передатчика (с 0)
} T_INFO_CHANNEL;

/// /// /// /// ///
/// \remark DATA - Работа с прерываниями
/// /// /// /// ///

/// \brief управление слежением за РК
typedef struct {
    unsigned int manage;                        ///< вкл/выкл слежение
    unsigned int startTrackingFreeTimer;        ///< начальное значение freetimer
} T_TRACKING_INT;

/// \brief выключить слежение за РК
#define T_TRACKING_INT_MAN_OFF      0
/// \brief включить слежение за РК
#define T_TRACKING_INT_MAN_ON       ~T_TRACKING_INT_MAN_OFF

/// \brief кол-во прерываний РК
#define T_CHECK_TRACKING_INT_CNT    8

/// \brief номер входной РК 1
#define T_CHECK_TRACKING_INT_RKIN_1 0
/// \brief номер входной РК 2
#define T_CHECK_TRACKING_INT_RKIN_2 1
/// \brief номер входной РК 3
#define T_CHECK_TRACKING_INT_RKIN_3 2
/// \brief номер входной РК 4
#define T_CHECK_TRACKING_INT_RKIN_4 3
/// \brief номер входной РК 5
#define T_CHECK_TRACKING_INT_RKIN_5 4
/// \brief номер входной РК 6
#define T_CHECK_TRACKING_INT_RKIN_6 5
/// \brief номер входной РК 7
#define T_CHECK_TRACKING_INT_RKIN_7 6
/// \brief номер входной РК 8
#define T_CHECK_TRACKING_INT_RKIN_8 7

/// \brief состояние рк вход
typedef struct {
    unsigned int startTrackingFreeTimer;        ///< начальное значение freetimer
    unsigned int finishTrackingFreeTimer;       ///< начальное значение freetimer
    unsigned int rkInAppearCount[T_CHECK_TRACKING_INT_CNT];            ///< кол-во возникновений РК Вход
} T_CHECK_TRACKING_INT;

/// \brief входной РК - ничего не выбрано
#define RKIN_NONE      0
/// \brief номер входная РК 1
#define RKIN_NUM_1     1
/// \brief номер входная РК 2
#define RKIN_NUM_2     2
/// \brief номер входная РК 3
#define RKIN_NUM_3     3
/// \brief номер входная РК 4
#define RKIN_NUM_4     4
/// \brief номер входная РК 5
#define RKIN_NUM_5     5
/// \brief номер входная РК 6
#define RKIN_NUM_6     6
/// \brief номер входная РК 7
#define RKIN_NUM_7     7
/// \brief номер входная РК 8
#define RKIN_NUM_8     8

/// \brief нижний фронт входной РК
#define FRONT_DOWN      0
/// \brief верхний фронт входной РК
#define FRONT_UP        ~FRONT_DOWN

/***************************************************************************************************/

/// /// /// /// /// /// /// /// /// ///
/// \remark КОДЫ ОШИБОК IOCTL-запросов
/// /// /// /// /// /// /// /// /// ///

/// \brief нет ошибок
#define GOOD                            0
/// \brief ошибка
#define ERROR                           -1000
/// \brief ошибка адреса регистра
#define ERR_BAD_ADDRESS                 -1001
/// \brief ошибка копирования данных в юзерспейс
#define ERR_COPY_TO_USER                -1003
/// \brief ошибка аргумента - typr
#define ERR_BAD_ARG_TYPE                -1011
/// \brief ошибка выделения памяти
/// внутренняя ошибка драйвера
#define ERR_BAD_ALLOC_MEM               -1012
/// \brief ошибка - невалидный дма индекс
#define ERR_BAD_DMA_INDEX               -1013
/// \brief ошибка - невалидный номер дма блока
#define ERR_BAD_DMA_NBLOCK              -1014


#endif // MIL1553UD_CMD_H
