/// \file a429_ioctl.с
/// \brief инфраструктура для обработки ioctl команд
#include "a429_ioctl.h"
#include "a429_log.h"
#include "a429_regs.h"
#include "a429_info.h"
#include "a429_cdev.h"
#include <linux/timekeeping.h>

/// \remark заголовки функций ioctl-команд

long ioctl_write_reg(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_write_reg_bit_mask(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_read_reg(struct a429_cdev_data* dev, unsigned long arg);

long ioctl_read_dma_blocks(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_clear_dma(struct a429_cdev_data* dev, unsigned long arg);

long ioctl_version(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_version_driver(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_pci_location(struct a429_cdev_data* dev, unsigned long arg);

long ioctl_write_to_ram(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_read_from_ram(struct a429_cdev_data* dev, unsigned long arg);

long ioctl_info_channel(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_info_device(struct a429_cdev_data* dev, unsigned long arg);

long ioctl_tracking_int(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_check_tracking_int(struct a429_cdev_data* dev, unsigned long arg);
long ioctl_start_as_rk_in(struct a429_cdev_data* dev, unsigned long arg);

/// \remark реализация корневого ioctl-обработчика

long ioctl_handler(struct a429_cdev_data* dev, unsigned int cmd, unsigned long arg)
{
    long err = 0;
    if (_IOC_TYPE( cmd ) != IOC_MAGIC) {
        LOG_ERROR("ioctl_handler - bad IOC_MAGIC");
        return -ENOTTY;
    }
    LOG_INFO("devIndex = %d, chNum = %d, drvIndex = %d", dev->index, dev->numCh, dev->drv->unit);
    switch (cmd) {
    case IOCTL_WRITE_REG:                       err = ioctl_write_reg(dev, arg);
        break;
    case IOCTL_WRITE_REG_BIT_MASK:              err = ioctl_write_reg_bit_mask(dev, arg);
        break;
    case IOCTL_READ_REG:                        err = ioctl_read_reg(dev, arg);
        break;
    case IOCTL_READ_DMA:                        err = ioctl_read_dma_blocks(dev, arg);
        break;
    case IOCTL_CLEAR_DMA:                       err = ioctl_clear_dma(dev, arg);
        break;
    case IOCTL_VERSION:                         err = ioctl_version(dev, arg);
        break;
    case IOCTL_VERSION_DRIVER:                  err = ioctl_version_driver(dev, arg);
        break;
    case IOCTL_PCI_LOCATION:                    err = ioctl_pci_location(dev, arg);
        break;
    case IOCTL_WRITE_TO_RAM:                    err = ioctl_write_to_ram(dev, arg);
        break;
    case IOCTL_READ_FROM_RAM:                   err = ioctl_read_from_ram(dev, arg);
        break;
    case IOCTL_INFO_CHANNEL:                    err = ioctl_info_channel(dev, arg);
        break;
    case IOCTL_INFO_DEVICE:                     err = ioctl_info_device(dev, arg);
        break;
    case IOCTL_TRACKING_INT:                    err = ioctl_tracking_int(dev, arg);
        break;
    case IOCTL_CHECK_TRACKING_INT:              err = ioctl_check_tracking_int(dev, arg);
        break;
    case IOCTL_START_AS_RK_IN:                  err = ioctl_start_as_rk_in(dev, arg);
        break;

    default:
        LOG_ERROR("ioctl_handler - understand cmd - %d", cmd);
        return -ENOTTY;
    }
    // обработка ошибок
    if (err)
        LOG_ERROR("ioctl_handler - error - %d", err);
    return err;
}

/// \remark реализация ioctl-команд

long ioctl_write_reg(struct a429_cdev_data* dev, unsigned long arg)
{
    SADDR_DATA data;
    LOG_INFO("ioctl_write_reg");
    copy_from_user((void *) &data, (void *) arg, sizeof(SADDR_DATA));
    spin_lock(&dev->spinlockCh);
    iowrite32(data.data, dev->drv->pbar + data.daddr);
    spin_unlock(&dev->spinlockCh);
    return GOOD;
}

/// \brief макрос записи заданной битовой маски в регистр
#define SET_REG_BIT_MASK(dev, data, regVal, shift)  \
    regVal = ioread32(dev->drv->pbar + data.daddr); /* считываем значение регистра */ \
    for(shift = 0; shift < 32; shift++) { /* обходим все биты */ \
        if (data.mask & (1 << shift)) { /* 1 - изменяем бит */ \
            if (data.data & (1 << shift)) /* смотрим исходное значение бита */ \
                regVal |= (1 << shift); /* устанавливаем 1 */ \
            else \
                regVal &= ~(1 << shift); /* устанавливаем 0 */ \
        } else { /* 0 - не изменяем бит */ \
            continue; \
        } \
    }

long ioctl_write_reg_bit_mask(struct a429_cdev_data* dev, unsigned long arg)
{
    SADDR_DATA_BIT_MASK data;   // исходные данные
    unsigned int regVal;        // временное значение регистра
    int shift;                  // смещение бита
    LOG_INFO("ioctl_write_reg_bit_mask");
    copy_from_user((void *) &data, (void *) arg, sizeof(SADDR_DATA_BIT_MASK));
    spin_lock(&dev->spinlockCh);
    SET_REG_BIT_MASK(dev, data, regVal, shift);
    iowrite32(regVal, dev->drv->pbar + data.daddr);
    spin_unlock(&dev->spinlockCh);
    return GOOD;
}

long ioctl_read_reg(struct a429_cdev_data* dev, unsigned long arg)
{
    SADDR_DATA data;
    LOG_INFO("ioctl_read_reg");
    copy_from_user((void *) &data, (void *) arg, sizeof(SADDR_DATA));
    spin_lock(&dev->spinlockCh);
    data.data = ioread32(dev->drv->pbar + data.daddr);
    spin_unlock(&dev->spinlockCh);
    if (copy_to_user((void*)arg, (void*)&data, sizeof(SADDR_DATA)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

/// \brief адрес укзателя записи дма для канала
#define DMA_WR_ADR            dev->dma_wr
/// \brief адрес укзателя чтения дма для канала
#define DMA_RD_ADR            dev->dma_rd
/// \brief базовый адрес указателя буфера ДМА
#define DMA_VADR              dev->drv->dma_info.dma_vaddr
/// \brief значение базового адреса указателя буфера ДМА
#define VALADR_DMA_VADR       (uint64_t)DMA_VADR
/// \brief значение адреса конца буфера дма для канала
#define VALADR_DMA_CH_END     (VALADR_DMA_VADR + DMA_SIZE)
/// \brief значение адреса начала буфера дма для канала
#define VALADR_DMA_CH_BEGIN   (VALADR_DMA_VADR)
/// \brief значение адреса укзателя записи дма для канала
#define VALADR_DMA_WR         (uint64_t)DMA_WR_ADR
/// \brief значение адреса укзателя чтения дма для канала
#define VALADR_DMA_RD         (uint64_t)DMA_RD_ADR


long ioctl_read_dma_blocks(struct a429_cdev_data* dev, unsigned long arg)
{
    T_CONTAINER_DMA* block;                 // контейнер считываемых блоков dma
    uint32_t dmaIndex = 0;                  // индекс dma канала
    int readyReadBlocksCount = -1;          // кол-во блоков, готовых для чтения
    T_INFO_CHANNEL info;                    // информация о канале
    int indexContainerBlock = 0;            // индекс блока в контейнере
    T_BLOCK_DMA* blockDma;                  // ссылка для работы с очередным блоком в дма
    spin_lock(&dev->spinlockCh);
    // получаем информацию о текущем канале
    info = get_info_channel(dev);
    LOG_INFO("ioctl_read_dma_blocks ch = %d", info.numberChannel);
    if ((info.typeChannel == T_INFO_CHANNEL_BAD_VALUE) || (info.numberChannel == T_INFO_CHANNEL_BAD_VALUE)) {
        spin_unlock(&dev->spinlockCh);
        return -ERR_BAD_ARG_TYPE;
    }
    // чтение индекса dma канала
    dmaIndex = ioread32(dev->drv->pbar + REG_DMA_INDEX);    
    if (dmaIndex == 0xFFFFFFFF) {
        LOG_INFO("%s bad dma_index_reg = 0x%x", dev->devname, dmaIndex);
        spin_unlock(&dev->spinlockCh);
        return -EFAULT;
    }
    // нормируем значение индекса дма
    LOG_INFO("dma index = 0x%x ch = %d", dmaIndex, info.numberChannel);
    dmaIndex = dmaIndex >> 4;    
    // обновляем указатель записи
    DMA_WR_ADR  = (uint32_t*)(DMA_VADR + dmaIndex * DMA_RAW_BLOCK_SIZE);
    // считаем кол-во готовых к чтению блоков
    if ( VALADR_DMA_WR >= VALADR_DMA_RD ) { // не превышает границу буфера
        readyReadBlocksCount = (VALADR_DMA_WR - VALADR_DMA_RD) / DMA_RAW_BLOCK_SIZE;
        LOG_INFO("yes - readyReadBlocksCount = 0x%d ch = %d", readyReadBlocksCount, info.numberChannel);
    } else {
        readyReadBlocksCount = (VALADR_DMA_WR - VALADR_DMA_CH_BEGIN) / DMA_RAW_BLOCK_SIZE + \
                (VALADR_DMA_CH_END - VALADR_DMA_RD) / DMA_RAW_BLOCK_SIZE;
        LOG_INFO("no - readyReadBlocksCount = 0x%d ch = %d", readyReadBlocksCount, info.numberChannel);
    }
    if (readyReadBlocksCount == 0) {
        copy_to_user((void*)(arg + sizeof(int)), (void*)&readyReadBlocksCount, sizeof(int));
        spin_unlock(&dev->spinlockCh);
        return 0;
    }
    block = kzalloc(sizeof(T_CONTAINER_DMA), GFP_KERNEL);
    block->count = 0;
    // копируем по одному блоку с проверкой на принадлежность к каналу
    while ((readyReadBlocksCount > 0) && (block->count < DMA_COUNT_BLOCKS)) {
        if (VALADR_DMA_RD >= VALADR_DMA_CH_END)  // проверяем переход через край кольцевого буфера
            DMA_RD_ADR = (uint32_t*)(DMA_VADR);
        blockDma = (T_BLOCK_DMA*)DMA_RD_ADR;
        // проверяем очередной блок в дма на принадлежность к каналу
        if ((info.typeChannel == T_BLOCK_DMA_TYPE_CH(blockDma->word1)) &&
            info.numberChannel == T_BLOCK_DMA_NUM_CH(blockDma->word1)) {
            block->blocks[indexContainerBlock++] = *blockDma; // копируем очередной блок в контейнер
            block->count++; // увеличиваем кол-во блоков в контейнере
        }
        if (T_BLOCK_DMA_TYPE_CH(blockDma->word1))
            LOG_INFO("dma rd= 0x%x, w1= 0x%x, w2= 0x%x, w3= 0x%x, w4= 0x%x, tCh= trmt(%d), nCh= %d", DMA_RD_ADR,
                     blockDma->word1, blockDma->word2, blockDma->word3, blockDma->word4,
                     T_BLOCK_DMA_TYPE_CH(blockDma->word1), T_BLOCK_DMA_NUM_CH(blockDma->word1));
        else
            LOG_INFO("dma rd= 0x%x, w1= 0x%x, w2= 0x%x, w3= 0x%x, w4= 0x%x, tCh= rcv(%d), nCh= %d", DMA_RD_ADR,
                     blockDma->word1, blockDma->word2, blockDma->word3, blockDma->word4,
                     T_BLOCK_DMA_TYPE_CH(blockDma->word1), T_BLOCK_DMA_NUM_CH(blockDma->word1));
        readyReadBlocksCount--; // учитываем проверенный блок
        DMA_RD_ADR += DMA_RAW_BLOCK_SIZE / 4; // сдвигаем указатель чтения на один проверенный блок
    }
    if (info.typeChannel)
        LOG_INFO("%s trmt= %d curDmaIndex = %d, readyReadBlocksCount = %d, countBlocks = %d",
                dev->devname, info.numberChannel, dmaIndex, readyReadBlocksCount, block->count);
    else
        LOG_INFO("%s rcv= %d curDmaIndex = %d, readyReadBlocksCount = %d, countBlocks = %d",
                dev->devname, info.numberChannel, dmaIndex, readyReadBlocksCount, block->count);
    copy_to_user((void*)arg, (void*)block, sizeof(T_CONTAINER_DMA));  // копируем в юзерспейc
    kfree(block);
    spin_unlock(&dev->spinlockCh);
    return GOOD;
}

long ioctl_clear_dma(struct a429_cdev_data* dev, unsigned long arg)
{
    int index = 0;
    spin_lock(&dev->spinlockCh);
    iowrite32(0, dev->drv->pbar + REG_DMA_INDEX);
    for (index = 0; index < (dev->drv->curCountReceiveChannels+dev->drv->curCountTransmitChannels); index++){
        getCdevPtrByIndex(dev->drv->index_cdev[index])->dma_wr = (unsigned int*)dev->drv->dma_info.dma_vaddr;
        getCdevPtrByIndex(dev->drv->index_cdev[index])->dma_rd = (unsigned int*)dev->drv->dma_info.dma_vaddr;
    }
    spin_unlock(&dev->spinlockCh);
    return GOOD;
}

long ioctl_version(struct a429_cdev_data* dev, unsigned long arg)
{
    int i;              // счётчик
    VERSION version;    // подробная информация о драйвере и устройстве
    LOG_INFO("ioctl_version");
    version.vendor_id = dev->drv->vendorid;
    version.device_id = dev->drv->deviceid;
    version.revision = dev->drv->revisionid;
    version.type = 0;
    for (i = 0; i < 30; i++)
        version.dev_name[i] = dev->drv->devname[i];
    version.minor = dev->numCh;
    version.irq = 0; // поправить
    version.size_dma = dev->drv->dma_info.dma_size;
    version.addr_dma_virt = dev->drv->dma_info.dma_vaddr;
    version.pci_bars = dev->drv->bar;
    version.addr_bar_virt = dev->drv->pbar;
    if (copy_to_user((void*)arg, (void*)&version, sizeof(VERSION)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

long ioctl_version_driver(struct a429_cdev_data* dev, unsigned long arg)
{
    unsigned int driverVersion;     // версия драйвера
    LOG_INFO("ioctl_version_driver");
    driverVersion = DRIVER_DATE_N_VERSION;
    if (copy_to_user((void*)arg, (void*)&driverVersion, sizeof(unsigned int)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

long ioctl_pci_location(struct a429_cdev_data* dev, unsigned long arg)
{
    int i;                      // счётчик
    PCI_LOCATION pciLoc;        // информация о pci-локации устройства
    LOG_INFO("ioctl_pci_location");
    for ( i = 0; i < SIZE_OF_PCI_LOC; i++)
        pciLoc.name[i] = 0;
    snprintf(pciLoc.name, SIZE_OF_PCI_LOC, "%s." , pci_name(dev->drv->pdevice));
    if (copy_to_user((void*)arg, (void*)&pciLoc, sizeof(PCI_LOCATION)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

long ioctl_write_to_ram(struct a429_cdev_data* dev, unsigned long arg)
{
    T_INFO_CHANNEL info;
    RAM_BLOCK block;                // данные из юзерспейс
    unsigned int regAddr;           // адрес регистра
    unsigned int index;             // текущий индекс слова
    unsigned int* dWords;           // инструкции
    LOG_INFO("ioctl_write_to_ram");
    spin_lock(&dev->spinlockCh);
    info = get_info_channel(dev);
    copy_from_user((void *) &block, (void *) arg, sizeof(RAM_BLOCK)); // получаем данные из юзерспейс
    // выбрать адрес куда писать в зависимости от типа
    switch (block.type) {
    case RAM_BLOCK_TYPE_DATA:           regAddr = REG_TX_DATA_RAM(info.numberChannel);
        break;
    case RAM_BLOCK_TYPE_DESCRIPTOR:     regAddr = REG_TX_DESC_RAM(info.numberChannel);
        break;
    default:
        LOG_ERROR("bad argument - type = %d", block.type);
        spin_unlock(&dev->spinlockCh);
        return ERR_BAD_ARG_TYPE;
    }
    // выделяем память для данных из юзерспейс
    dWords = kmalloc(block.length * 4, GFP_KERNEL);
    if (!dWords) {
        LOG_ERROR("bad alloc mem for dwords");
        spin_unlock(&dev->spinlockCh);
        return ERR_BAD_ALLOC_MEM;
    }
    // копируем инструкции из юзерспейс
    copy_from_user((void *) dWords, (void *) (block.dwords), block.length * 4);
    // добавляем смещенние
    regAddr += block.shift * 4;
    // пословная запись слов в ram
    for (index = 0; index < block.length; index++) {
        iowrite32(dWords[index], dev->drv->pbar + regAddr);
        regAddr += 4;
        //LOG_INFO("wrt word = 0x%x at 0x%x", dWords[index], regAddr + (block.shift + index) * 4);
    }
    kfree(dWords);
    spin_unlock(&dev->spinlockCh);
    return GOOD;
}

long ioctl_read_from_ram(struct a429_cdev_data* dev, unsigned long arg)
{
    T_INFO_CHANNEL info;
    RAM_BLOCK block;             // данные из юзерспейс
    unsigned int regAddr;           // адрес регистра
    unsigned int index;             // текущий индекс слова
    unsigned int* dWords;           // инструкции
    LOG_INFO("ioctl_read_from_ram");
    spin_lock(&dev->spinlockCh);
    info = get_info_channel(dev);
    copy_from_user((void *) &block, (void *) arg, sizeof(RAM_BLOCK)); // получаем данные из юзерспейс
    // выбрать адрес откуда читать в зависимости от типа
    switch (block.type) {
    case RAM_BLOCK_TYPE_DATA:           regAddr = REG_TX_DATA_RAM(info.numberChannel);
        break;
    case RAM_BLOCK_TYPE_DESCRIPTOR:     regAddr = REG_TX_DESC_RAM(info.numberChannel);
        break;
    default:
        LOG_ERROR("bad argument - type = %d", block.type);
        spin_unlock(&dev->spinlockCh);
        return ERR_BAD_ARG_TYPE;
    }
    // выделяем память для данных из юзерспейс
    dWords = kmalloc(block.length * 4, GFP_KERNEL);
    if (!dWords) {
        LOG_ERROR("bad alloc mem for dwords");
        spin_unlock(&dev->spinlockCh);
        return ERR_BAD_ALLOC_MEM;
    }
    // добавляем смещенние
    regAddr += block.shift * 4;
    // пословное чтение слов из ram
    for (index = 0; index < block.length; index++) {
        dWords[index] = ioread32(dev->drv->pbar + regAddr);
        regAddr += 4;
        //LOG_INFO("rd word = 0x%x at 0x%x", dWords[index], regAddr + (block.shift + index) * 4);
    }
    // копируем в юзерспейс блок
    copy_to_user((void*)block.dwords, (void*)dWords, block.length * 4);
    kfree(dWords);
    spin_unlock(&dev->spinlockCh);
    return GOOD;
}

long ioctl_info_channel(struct a429_cdev_data* dev, unsigned long arg) {
    T_INFO_CHANNEL info;
    LOG_INFO("ioctl_info_channel");
    spin_lock(&dev->spinlockCh);
    copy_from_user((void *) &info, (void *) arg, sizeof(T_INFO_CHANNEL));
    info = get_info_channel(dev);
    copy_to_user((void*)arg, (void*)&info, sizeof(T_INFO_CHANNEL));
    if ((info.typeChannel == T_INFO_CHANNEL_BAD_VALUE) || (info.numberChannel == T_INFO_CHANNEL_BAD_VALUE)) {
        spin_unlock(&dev->spinlockCh);
        return -ERR_BAD_ARG_TYPE;
    }
    spin_unlock(&dev->spinlockCh);
    return GOOD;
}

long ioctl_info_device(struct a429_cdev_data* dev, unsigned long arg) {
    T_INFO_DEVICE info;
    LOG_INFO("ioctl_info_device");
    spin_lock(&dev->spinlockCh);
    info.countRecieverChannels = dev->drv->curCountReceiveChannels;
    info.countTransmitterChannels = dev->drv->curCountTransmitChannels;
    copy_to_user((void*)arg, (void*)&info, sizeof(T_INFO_DEVICE));
    spin_unlock(&dev->spinlockCh);
    return GOOD;
}

long ioctl_tracking_int(struct a429_cdev_data* dev, unsigned long arg)
{
    struct timespec64 tv;
    int i;
    T_TRACKING_INT trackingInt;
    ktime_get_real_ts64(&tv);
    spin_lock_irqsave(&dev->drv->spinlockIrq, dev->drv->spinlockIrqFlags);
    copy_from_user((void *) &trackingInt, (void *) arg, sizeof(T_TRACKING_INT));
    dev->drv->trackingInt.manage = trackingInt.manage;
    trackingInt.startTrackingFreeTimer = tv.tv_nsec;
    dev->drv->trackingInt.startTrackingFreeTimer = trackingInt.startTrackingFreeTimer;
    copy_to_user((void*)arg, (void*)&trackingInt, sizeof(T_TRACKING_INT));
    // подготовка буфера результатов
    dev->drv->checkTrackingInt.startTrackingFreeTimer = trackingInt.startTrackingFreeTimer;
    dev->drv->checkTrackingInt.finishTrackingFreeTimer = 0;
    for (i=0; i<T_CHECK_TRACKING_INT_CNT; i++){
        dev->drv->checkTrackingInt.rkInAppearCount[i] = 0;
    }
    spin_unlock_irqrestore(&dev->drv->spinlockIrq, dev->drv->spinlockIrqFlags);
    return GOOD;
}

long ioctl_check_tracking_int(struct a429_cdev_data* dev, unsigned long arg)
{
    struct timespec64 tv;
    ktime_get_real_ts64(&tv);
    spin_lock_irqsave(&dev->drv->spinlockIrq, dev->drv->spinlockIrqFlags);
    dev->drv->checkTrackingInt.finishTrackingFreeTimer = tv.tv_nsec;
    copy_to_user((void*)arg, (void*)&dev->drv->checkTrackingInt, sizeof(T_CHECK_TRACKING_INT));
    spin_unlock_irqrestore(&dev->drv->spinlockIrq, dev->drv->spinlockIrqFlags);
    return GOOD;
}

long ioctl_start_as_rk_in(struct a429_cdev_data* dev, unsigned long arg)
{
    unsigned int numRkIn;
    spin_lock_irqsave(&dev->drv->spinlockIrq, dev->drv->spinlockIrqFlags);
    copy_from_user((void *) &numRkIn, (void *) arg, sizeof(unsigned int));
    dev->numRkIn = numRkIn;
    spin_unlock_irqrestore(&dev->drv->spinlockIrq, dev->drv->spinlockIrqFlags);
    return GOOD;
}

