/// \file a429_pci.с
/// \brief инфраструктура для работы с pci устройством (pci-драйвер)
#include "a429_pci.h"
#include "a429_log.h"
#include "a429_regs.h"
#include "a429_cdev.h"
#include "a429_ioctl.h"

/// \brief инициализация pci-драйвера
static int __init a429_pci_init(void);
/// \brief деинициализация pci-драйвера
static void __exit a429_pci_exit(void);
/// \brief открытие (инициализация) очередного pci-устройства
static int a429_pci_probe(struct pci_dev *pdevice, const struct pci_device_id *pid);
/// \brief закрытие (деинициализация) очередного pci-устройства
static void a429_pci_remove(struct pci_dev *pdevice);
/// \brief обработчик прерывания pci-устройства
static irqreturn_t a429_pci_interrupt_handler(int irq, void* arg);
/// \brief инициализация кол-ва каналов pci-устройства
static void init_count_of_channels(struct A429_DEV* driver_priv);
/// \brief разовый запуск передатчика
static void start_once(struct a429_cdev_data* dev);
/// \brief включение приёмника
static void on_receiver(struct a429_cdev_data* dev);
/// \brief реализация функциональности свободных вх рк
static void a429_pci_interrupt_handler_read_rk_in(struct A429_DEV* drv, unsigned int interrupt);
/// \brief реализация функциональности вх рк запрос на передачу и готовность к приёму
static void a429_pci_interrupt_handler_request_n_ready_rk_in(struct A429_DEV* drv, unsigned int interrupt);

/// \brief pdevs глобальный массив открытых pci-устройств
struct A429_DEV* pdevs[MAX_COUNT_BOARDS];

MODULE_LICENSE(DRIVER_LICENSE);
MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_VERSION(DRIVER_VERSION);

/// \brief регистрация функции инициализации pci-драйвера
module_init(a429_pci_init);
/// \brief регистрация функции деинициализации pci-драйвера
module_exit(a429_pci_exit);

/// \brief таблица pci-устройств
static struct pci_device_id a429_pci_id_table[] = {
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_PCIE_A429)   },
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_MPCIE_A429)  },
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_XMC_A429)  },
    { 0 }
};
MODULE_DEVICE_TABLE(pci, a429_pci_id_table);

/// \brief таблица операций драйвера pci-устройств
static struct pci_driver a429_pci_driver = {
    .name           = DRIVER_NAME,
    .id_table       = a429_pci_id_table,
    .probe          = a429_pci_probe,
    .remove         = a429_pci_remove,
};

static void init_count_of_channels(struct A429_DEV* driver_priv) {
    switch (driver_priv->revisionid) {
    case DEVICE_REV_ID_A429_UD88:{
        driver_priv->curCountReceiveChannels = 8;
        driver_priv->curCountTransmitChannels = 8;
    }break;
    case DEVICE_REV_ID_A429_UD84:{
        driver_priv->curCountReceiveChannels = 8;
        driver_priv->curCountTransmitChannels = 4;
    }break;
    case DEVICE_REV_ID_A429_UD44:{
        driver_priv->curCountReceiveChannels = 4;
        driver_priv->curCountTransmitChannels = 4;
    }break;
    case DEVICE_REV_ID_A429_UD66:{
        driver_priv->curCountReceiveChannels = 16;
        driver_priv->curCountTransmitChannels = 16;
    }break;
    case DEVICE_REV_ID_A429_UD22:{
        driver_priv->curCountReceiveChannels = 12;
        driver_priv->curCountTransmitChannels = 12;
    }break;
    default: LOG_ERROR("init_count_of_channels - can't identify deviceid"); break;
    }
    LOG_RELEASE_INFO("curCountReceiveChannels: %d, curCountTransmitChannels: %d", driver_priv->curCountReceiveChannels, driver_priv->curCountTransmitChannels);
}

static int __init a429_pci_init(void)
{
    int error, i;
    for (i = 0; i < MAX_COUNT_BOARDS; i++)
        pdevs[i] = NULL;
    clear_cdev_array();
    LOG_RELEASE_INFO("Init %s - version %s\n", DRIVER_DESCRIPTION, DRIVER_VERSION);
    LOG_INFO("%s\n", DRIVER_COPYRIGHT);
    error = pci_register_driver(&a429_pci_driver);
    if (error != 0) {
      LOG_ERROR("pci_register_driver error %d", error);
    }    
    LOG_RELEASE_INFO("Init End \n");
    return error;
}

static void __exit a429_pci_exit(void)
{
    LOG_RELEASE_INFO("Exit %s - version %s\n", DRIVER_DESCRIPTION, DRIVER_VERSION);   
    pci_unregister_driver(&a429_pci_driver);
}

static int a429_pci_probe(struct pci_dev* pdevice, const struct pci_device_id* pid)
{
    int bar, err, i;
    uint64_t mmio_start, mmio_len;
    uint16_t status_reg, command_reg;
    struct A429_DEV* driver_priv;
    u32 addr_dma_lo, addr_dma_hi;
    // проверка допустимости девайса по вендору поддерживаемого устройства
    if (pdevice->vendor != VENDOR_ID_NOVOMAR)
        return -ENXIO;
    // проверка допустимости девайса по идентификатору поддерживаемого устройства
    if ((pdevice->device != DEVICE_ID_PCIE_A429) &&
        (pdevice->device != DEVICE_ID_XMC_A429) &&
        (pdevice->device != DEVICE_ID_MPCIE_A429))
        return -ENXIO;
    // проверка допустимости девайса по ревизии поддерживаемого устройства
    if ((pdevice->revision != DEVICE_REV_ID_A429_UD88) &&
        (pdevice->revision != DEVICE_REV_ID_A429_UD84) &&
        (pdevice->revision != DEVICE_REV_ID_A429_UD44) &&
        (pdevice->revision != DEVICE_REV_ID_A429_UD22) &&
        (pdevice->revision != DEVICE_REV_ID_A429_UD66))
        return -ENXIO;
    // инициализация вспомогательной структуры данных PCI-драйвера
    driver_priv = kzalloc(sizeof(struct A429_DEV), GFP_KERNEL);
    pci_set_drvdata(pdevice, driver_priv);
    driver_priv->pdevice = pdevice;
    // инициализация регистров управления a429
    bar = pci_select_bars(pdevice, IORESOURCE_MEM); // получаем адрес реcурса ввода/вывода
    LOG_RELEASE_INFO("Device availale MEM BAR are 0x%x\n", bar);
    err = pci_enable_device(pdevice); // "включаем" память устройства
    if (err != 0) {
      LOG_ERROR("pci_enable_device_mem error %d", err);
      return err;
    }
    pci_set_master(pdevice);
    err = pci_request_region(pdevice, bar, DRIVER_NAME); // запрашиваем необходимый регион пямяти, с определённым ранее типом
    if (err != 0) {
      LOG_ERROR("pci_request_region error %d", err);
      return err;
    }
    driver_priv->bar = bar;
    mmio_start = pci_resource_start(pdevice, 0); // получаем адрес начала блока памяти устройства
    mmio_len = pci_resource_len(pdevice, 0); // получаем общую длину блока памяти устройства
    LOG_RELEASE_INFO("a429 device resource 0: start at 0x%lx with lenght %lu\n", mmio_start, mmio_len);
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 2, 0)
		driver_priv->pbar = ioremap_cache(mmio_start, mmio_len); // мапим выделенную память к аппаратуре
	#else
		driver_priv->pbar = ioremap_nocache(mmio_start, mmio_len); // мапим выделенную память к аппаратуре
	#endif
    driver_priv->barlen = mmio_len; // сохраняем длину выделенной памяти
    driver_priv->pdevice = pdevice; // сохраняем указатель на pci_dev
    LOG_RELEASE_INFO("a429 device mapped resource 0x%lx to 0x%p\n", mmio_start, driver_priv->pbar);
    // читаем конфигурационные регистры PCI
    pci_read_config_word(pdevice, PCI_VENDOR_ID, &driver_priv->vendorid);
    pci_read_config_word(pdevice, PCI_DEVICE_ID, &driver_priv->deviceid);
    pci_read_config_word(pdevice, PCI_REVISION_ID, &driver_priv->revisionid);
    LOG_RELEASE_INFO("Device vid: 0x%X  pid: 0x%X revid: 0x%X\n", driver_priv->vendorid, driver_priv->deviceid, driver_priv->revisionid);
    pci_read_config_word(pdevice, PCI_STATUS, &status_reg);
    pci_read_config_word(pdevice, PCI_COMMAND, &command_reg);
    LOG_RELEASE_INFO("Device status reg: 0x%X device command reg: 0x%X\n", status_reg, command_reg);
    // сохраняем struct a429* driver_priv в массив pci устройств
    for (i = 0; i < MAX_COUNT_BOARDS; i++) {
        if (pdevs[i] == NULL) {
            pdevs[i] = driver_priv;
            driver_priv->unit = i;
            spin_lock_init(&driver_priv->spinlockIrq);
            LOG_RELEASE_INFO("Device number %d has added to list", i);
            break;
        }
    }
    //инициализация имени pci устройства
    snprintf(driver_priv->devname, sizeof(driver_priv->devname), "%s%u" , PCI_DEVICE_NAME_TEMPLATE, driver_priv->unit);
    #if LINUX_VERSION_CODE < KERNEL_VERSION(6, 0, 0)
    //set DMA mask
    if (pci_set_dma_mask(driver_priv->pdevice, DMA_BIT_MASK(64)) == 0) {
        pci_set_consistent_dma_mask(driver_priv->pdevice, DMA_BIT_MASK(64));
        LOG_RELEASE_INFO("Memory");
    }
    #endif
    // настройка DMA
    driver_priv->dma_info.dma_size = DMA_SIZE;
    driver_priv->dma_info.dma_vaddr = (uint8_t*)dma_alloc_coherent(&pdevice->dev, driver_priv->dma_info.dma_size, &driver_priv->dma_info.dma_paddr, GFP_KERNEL);
    if (driver_priv->dma_info.dma_vaddr == NULL) {
      LOG_ERROR("could not allocate %lu bytes for data buffer\n", (unsigned long)driver_priv->dma_info.dma_size);
      return -1;
    }
    memset(driver_priv->dma_info.dma_vaddr, 0, driver_priv->dma_info.dma_size);
    LOG_RELEASE_INFO("DMA_DATA_BASE phys@%llXH virt@%p\n", (unsigned long long)driver_priv->dma_info.dma_paddr, driver_priv->dma_info.dma_vaddr);

    addr_dma_lo = ((uint64_t)driver_priv->dma_info.dma_paddr & 0xffffffff);	// low part address DMA
    addr_dma_hi = (((uint64_t)driver_priv->dma_info.dma_paddr >> 32));	// hi part address DMA
    iowrite32(addr_dma_lo, (void*) (driver_priv->pbar + REG_DMA_DATA_BASE));
    iowrite32(addr_dma_hi, (void*) (driver_priv->pbar + REG_DMA_DATA_BASE + 0x4));

    // определение количества каналов на плате
    init_count_of_channels(driver_priv);
    // создание символьных устройств: на каждый канал - символное устройство
    create_char_devs(driver_priv);
    // инициализация прерывания
    disable_irq(pdevice->irq);  // надо сначала принудительно выключить - нашёл на stackoverflow
    err = pci_enable_msi(pdevice); // разрешаем прерывания
    if (err != 0) {
      LOG_ERROR("pci_enable_msi error %d", err);
      return err;
    }
    enable_irq(pdevice->irq); // включаем прерывания
    err = request_irq(pdevice->irq, a429_pci_interrupt_handler, IRQF_SHARED, driver_priv->devname, driver_priv); // привязываем обработчик прерывания
    if (err != 0) {
      LOG_ERROR("request_irq error %d", pdevice->irq);
      return err;
    }
    LOG_RELEASE_INFO("IRQ %d", pdevice->irq);
    LOG_RELEASE_INFO("probe finished!");
    return 0;
}

static void a429_pci_remove(struct pci_dev *pdevice)
{
    struct A429_DEV* driver_priv;
    driver_priv = pci_get_drvdata(pdevice);
    LOG_RELEASE_INFO("Deinit device %d", driver_priv->unit);
    // деинициализация ДМА
    dma_free_coherent(&pdevice->dev, driver_priv->dma_info.dma_size, driver_priv->dma_info.dma_vaddr, driver_priv->dma_info.dma_paddr);
    driver_priv->dma_info.dma_vaddr = NULL;
    driver_priv->dma_info.dma_paddr = 0;
    driver_priv->dma_info.dma_size = 0;
    // деинициализация прерывания
    free_irq(pdevice->irq, driver_priv);
    pci_disable_msi(pdevice);
    // уничтожение символьных устройств
    destroy_char_devs(driver_priv);
    // деинициализация вспомогательной структуры данных PCI-драйвера
    if (driver_priv) {
        if (driver_priv->pbar) {
            iounmap(driver_priv->pbar);
        }
        pdevs[driver_priv->unit] = NULL;
        kfree(driver_priv);
    }
    pci_release_region(pdevice, pci_select_bars(pdevice, IORESOURCE_MEM));
    pci_disable_device(pdevice);
    LOG_RELEASE_INFO("Unloaded %s\n", DRIVER_DESCRIPTION);
}

static irqreturn_t a429_pci_interrupt_handler(int irq, void* arg)
{
    unsigned int interrupt = 0;
    struct A429_DEV* driver_priv = (struct A429_DEV*)arg;     // информационная структура драйвера pci-устройства
    interrupt = ioread32(driver_priv->pbar + REG_INTERRUPT);
    spin_lock(&driver_priv->spinlockIrq);
    a429_pci_interrupt_handler_read_rk_in(driver_priv, interrupt);
    a429_pci_interrupt_handler_request_n_ready_rk_in(driver_priv, interrupt);;
    spin_unlock(&driver_priv->spinlockIrq);
    return IRQ_HANDLED;
}

#define A429_INT_SC1_MASK   8
#define A429_INT_SC2_MASK   9
#define A429_INT_SC3_MASK   10
#define A429_INT_SC4_MASK   11
#define A429_INT_SC5_MASK   12
#define A429_INT_SC6_MASK   13
#define A429_INT_SC7_MASK   14
#define A429_INT_SC8_MASK   15

#define A429_CHECK_TRACKING_INT_SC_PROC(num) \
    if (interrupt & (1 << A429_INT_SC##num##_MASK)) \
        drv->checkTrackingInt.rkInAppearCount[T_CHECK_TRACKING_INT_RKIN_##num]++

static void a429_pci_interrupt_handler_read_rk_in(struct A429_DEV* drv, unsigned int interrupt) {
    A429_CHECK_TRACKING_INT_SC_PROC(1);
    A429_CHECK_TRACKING_INT_SC_PROC(2);
    A429_CHECK_TRACKING_INT_SC_PROC(3);
    A429_CHECK_TRACKING_INT_SC_PROC(4);
    A429_CHECK_TRACKING_INT_SC_PROC(5);
    A429_CHECK_TRACKING_INT_SC_PROC(6);
    A429_CHECK_TRACKING_INT_SC_PROC(7);
    A429_CHECK_TRACKING_INT_SC_PROC(8);
    //printk("DEV=%s, devID=0x%x, revID=0x%x, interrupt= 0x%x \n", drv->devname, drv->deviceid, drv->revisionid, interrupt);
}

void a429_request_n_ready_int_sc_proc(unsigned int interrupt, int num, int i) {
//    printk("  interrupt= 0x%x, curnum= %d, mask= 0x%x \n", interrupt, num, 1 << (num + 7));
    if (interrupt & (1 << (num + 7))) {
//        printk("    rkinnum= %d == curnum= %d \n", getCdevPtrByIndex(i)->numRkIn, num);
        if (getCdevPtrByIndex(i)->numRkIn == num) {
//            printk("      yes - interrupt= 0x%x, rkinnum= %d \n", interrupt, getCdevPtrByIndex(i)->numRkIn);
            start_once(getCdevPtrByIndex(i));
            on_receiver(getCdevPtrByIndex(i));
        }
    }
}

static void a429_pci_interrupt_handler_request_n_ready_rk_in(struct A429_DEV* drv, unsigned int interrupt)
{
    int i;
    for (i=0; i<MAX_COUNT_BOARDS * (MAX_RECEIVE_CHANNELS + MAX_TRANSMIT_CHANNELS); i++) {
        if (getCdevPtrByIndex(i)->index == -1)
            continue;
        if ((unsigned long)drv == (unsigned long)getCdevPtrByIndex(i)->drv) {
//            printk("DEV=%s, numch= %d\n", drv->devname, i);
            a429_request_n_ready_int_sc_proc( interrupt, 1, i);
            a429_request_n_ready_int_sc_proc( interrupt, 2, i);
            a429_request_n_ready_int_sc_proc( interrupt, 3, i);
            a429_request_n_ready_int_sc_proc( interrupt, 4, i);
            a429_request_n_ready_int_sc_proc( interrupt, 5, i);
            a429_request_n_ready_int_sc_proc( interrupt, 6, i);
            a429_request_n_ready_int_sc_proc( interrupt, 7, i);
            a429_request_n_ready_int_sc_proc( interrupt, 8, i);
        }
    }
}

static void start_once(struct a429_cdev_data* dev)
{
    SADDR_DATA reg;
    T_INFO_CHANNEL chInfo;
    chInfo = get_info_channel(dev);
    if (chInfo.typeChannel == T_INFO_CHANNEL_TYPE_RECIEVER)
        return;
    reg.daddr = REG_TX_CONTROL(chInfo.numberChannel);
    reg.data = 0;
    reg.data = ioread32(dev->drv->pbar + reg.daddr);
    reg.data |= 1<<1;
    iowrite32(reg.data, dev->drv->pbar + reg.daddr);
}

static void on_receiver(struct a429_cdev_data* dev)
{
    SADDR_DATA reg;
    T_INFO_CHANNEL chInfo;
    chInfo = get_info_channel(dev);
    if (chInfo.typeChannel == T_INFO_CHANNEL_TYPE_TRANSMITTER)
        return;
    reg.daddr = REG_RX_CONF(chInfo.numberChannel);
    reg.data = 0;
    reg.data = ioread32(dev->drv->pbar + reg.daddr);
    reg.data |= 1<<31;
    iowrite32(reg.data, dev->drv->pbar + reg.daddr);
}


